{
/**********************************************************/
/*                      Chronos  V1.2                     */
/*                                                        */
/* Last release 01/05/2003                Sylvain Taufour */
/*                                sylvain.taufour@free.fr */
/**********************************************************/
}

{
/**********************************************************/
/*                        Historique                      */
/**********************************************************/
/*                                                        */
/* V1.2 : -Add a timer event (like TTimer)                */
/*                                                        */
/* V1.1 : -convert GetTickCount to QueryPerformanceCounter*/
/*         for best precision with Windows XP             */
/*         (thinks to DooMeeR)                            */
/*        -Add event                                      */
/**********************************************************/
}

unit Chronos;

interface

{$R Chronos.res}

uses
  Windows, Messages, SysUtils, Classes, Forms;

type
  TTimerThread = class(TThread)
  private
    FThInterval: Cardinal;
    FOnTimeToGo: TNotifyEvent;
    procedure ForSynchro;
  protected
    procedure Execute; override;
  public
    property ThInterval: Cardinal read FThInterval write FThInterval;
    property OnTimeToGo: TNotifyEvent read FOnTimeToGo write FOnTimeToGo;
  end;

  TWaitThread = class(TThread)
  private
    FDuration: Cardinal;
    FThStart: Cardinal;
    FNow: Cardinal;
  protected
    procedure Execute; override;
  public
    property Duration: Cardinal read FDuration write FDuration;
    property ThStart: Cardinal read FThStart write FThStart;
  end;

  TChronos = class(Tcomponent)
  private
    { Private declarations }
    FPastTime: Longword;
    FStart: Longword;
    FStop: Longword;
    FStopped: Boolean;
    FTimeBeforeEvent: Cardinal;
    FOnTimeIsGone: TNotifyEvent;
    FWThread: TWaitThread;
    FTThread: TTimerThread;
    FInterval: Cardinal;
    FEnable: Boolean;
    FOnTimer: TNotifyEvent;
    function GetPastTime: Longword;
    function GetStrPastTime: String;
    procedure SetEnable(const Value: Boolean);
    procedure SetOnTimer(const Value: TNotifyEvent);
    procedure SetInterval(const Value: Cardinal);
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Start;
    function Stop: Longword;
  published
    { Published declarations }
    property Enable: Boolean read FEnable write SetEnable;
    property Interval: Cardinal read FInterval write SetInterval;
    property PastTime: Longword read GetPastTime;
    property StrPastTime: String read GetStrPastTime;
    property TimeBeforeEvent: Cardinal read FTimeBeforeEvent write FTimeBeforeEvent;
    property OnTimeIsGone: TNotifyEvent read FOnTimeIsGone write FOnTimeIsGone;
    property OnTimer: TNotifyEvent read FOnTimer write SetOnTimer;
  end;

function MillisecondToStr(Ms: Longword): String;
function MyGetTickCount: Cardinal;
procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Samples', [TChronos]);
end;

{ TChronos }

constructor TChronos.Create(AOwner: TComponent);
begin
  inherited;

  FStart := 0;
  FStop := 0;
  FPastTime := 0;
  FInterval := 1000;
  FStopped := False;
  FEnable := False;
  FTimeBeforeEvent := 0;
end;

destructor TChronos.Destroy;
begin
  if Assigned(FWThread) then
    FWThread.Terminate;

  if Assigned(FTThread) then
    FTThread.Terminate;
    
  inherited;
end;

function TChronos.GetPastTime: Longword;
begin
  if FStopped then
    result := FStop - FStart
  else
    result := MyGetTickCount - FStart;
end;

function TChronos.GetStrPastTime: String;
begin
  result := MillisecondToStr(PastTime);
end;

procedure TChronos.SetEnable(const Value: Boolean);
begin
  FEnable := Value;
  if FEnable and not (csDesigning	in ComponentState) then
  begin
    if Assigned(FTThread) then
      FTThread.Terminate;

    FTThread := TTimerThread.Create(True);
    with FTThread do
    begin
      if Assigned(FOnTimer) then
        OnTimeToGo := FOnTimer;
        
      FreeOnTerminate := True;
      ThInterval := FInterval;
      Resume;
    end;
  end
  else 
    if Assigned(FTThread) then
    begin
      with FTThread do
      begin
        OnTimeToGo := nil;
        OnTerminate := nil;
        Terminate;
      end;
    end;
end;

procedure TChronos.SetInterval(const Value: Cardinal);
begin
  FInterval := Value;
  if Assigned(FTThread) then
    FTThread.FThInterval := FInterval;
end;

procedure TChronos.SetOnTimer(const Value: TNotifyEvent);
begin
  FOnTimer := Value;
  if Assigned(FTThread) then
    FTThread.OnTimeToGo := FOnTimer;
end;

procedure TChronos.Start;
begin
  FStart := MyGetTickCount;

  if FTimeBeforeEvent > 0 then
  begin
    if Assigned(FWThread) then
    begin
      FWThread.OnTerminate := nil;
      FWThread.Terminate;
    end;

    FWThread := TWaitThread.Create(True);
    with FWThread do
    begin
      FreeOnTerminate := True;
      if Assigned(FOnTimeIsGone) then
        OnTerminate := FOnTimeIsGone;
      ThStart := FStart;
      Duration := FTimeBeforeEvent;
      Resume;
    end;
  end;
end;

function TChronos.Stop: Longword;
begin
  result := FStop;
  if not FStopped then
  begin
    FStopped := True;
    FStop := MyGetTickCount;
    result := FStop - FStart;
  end;
end;

{ TWaitThread }

procedure TWaitThread.Execute;
begin
  FNow := MyGetTickCount;
  while not Terminated and (FNow - FThStart < FDuration) do
  begin
    FNow := MyGetTickCount;
  end;
end;

{ Functions }

function MillisecondToStr(Ms: Longword): String;
var
  H, Min, Sec, M: Integer;
begin
  result := '';
  M := Ms mod 1000;
  Ms := Ms div 1000;
  Sec := Ms mod 60;
  Min := (Ms mod 3600) div 60;
  H := Ms div 3600;
  result := Format('%.2d:%.2d:%.2d:%.3d', [H, Min, Sec, M]);
end;

function MyGetTickCount: Cardinal;
var
  lpPerformanceCount,
  lpFrequency: TLargeInteger;
begin
  if not QueryPerformanceCounter(lpPerformanceCount) then lpPerformanceCount := GetTickCount else
  begin
    QueryPerformanceFrequency(lpFrequency);
    lpPerformanceCount := (lpPerformanceCount * 1000) div lpFrequency
  end;

  result := lpPerformanceCount;
end;

{ TTimerThread }

procedure TTimerThread.Execute;
var
  FStart: Cardinal;
  FNow: Cardinal;
begin
  FStart := MyGetTickCount;
  while not Terminated do
  begin
    FNow := MyGetTickCount;
    if (FNow - FStart) >= FThInterval then
    begin
      if Assigned(FOnTimeToGo) then
        Synchronize(ForSynchro);

      FStart := MyGetTickCount;
    end;
  end;
end;

procedure TTimerThread.ForSynchro;
begin
  FOnTimeToGo(Self);
end;

end.
